/*
 * @(#)OGLMaskFill.java	1.10 06/06/27
 *
 * Copyright 2006 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package sun.java2d.opengl;

import java.awt.AlphaComposite;
import java.awt.Composite;
import java.awt.GradientPaint;
import java.awt.TexturePaint;
import sun.java2d.SunGraphics2D;
import sun.java2d.SurfaceData;
import sun.java2d.loops.GraphicsPrimitive;
import sun.java2d.loops.GraphicsPrimitiveMgr;
import sun.java2d.loops.CompositeType;
import sun.java2d.loops.MaskFill;
import sun.java2d.loops.SurfaceType;
import sun.java2d.pipe.Region;
import sun.java2d.pipe.RenderBuffer;
import sun.java2d.pipe.RenderQueue;
import static sun.java2d.loops.CompositeType.*;
import static sun.java2d.loops.SurfaceType.*;
import static sun.java2d.pipe.BufferedOpCodes.*;

/**
 * The MaskFill operation is expressed as:
 *   dst = ((src <MODE> dst) * pathA) + (dst * (1 - pathA))
 *
 * The OGL implementation of the MaskFill operation differs from the above
 * equation because it is not possible to perform such a complex operation in
 * OpenGL (without the use of advanced techniques like fragment shaders and
 * multitexturing).  Therefore, the OGLMaskFill operation is expressed as:
 *   dst = (src * pathA) <SrcOver> dst
 *
 * This simplified formula is only equivalent to the "true" MaskFill equation
 * in the following situations:
 *   - <MODE> is SrcOver
 *   - <MODE> is Src, extra alpha == 1.0, and the source paint is opaque
 *
 * Therefore, we register OGLMaskFill primitives for only the SurfaceType and
 * CompositeType restrictions mentioned above.  In addition, for the SrcNoEa
 * case we must override the incoming composite with a SrcOver (no extra alpha)
 * instance, so that we set up the OpenGL blending mode to match the
 * OGLMaskFill equation.
 */
abstract class OGLMaskFill extends MaskFill {

    static void register() {
        GraphicsPrimitive[] primitives = {
            new Solid(AnyColor, SrcOver),
            new Solid(OpaqueColor, SrcNoEa),

            new Gradient(GradientPaint, SrcOver),
            new Gradient(OpaqueGradientPaint, SrcNoEa),

            new Texture(TexturePaint, SrcOver),
            new Texture(OpaqueTexturePaint, SrcNoEa),
        };
        GraphicsPrimitiveMgr.register(primitives);
    }

    private final boolean isSimpleColor;

    protected OGLMaskFill(SurfaceType srcType,
                          CompositeType compType,
                          boolean isSimpleColor)
    {
        super(srcType, compType, OGLSurfaceData.OpenGLSurface);
        this.isSimpleColor = isSimpleColor;
    }

    @Override
    public void MaskFill(SunGraphics2D sg2d, SurfaceData sData,
                         Composite comp,
                         final int x, final int y, final int w, final int h,
                         final byte[] mask,
                         final int maskoff, final int maskscan)
    {
        AlphaComposite acomp = (AlphaComposite)comp;
        if (acomp.getRule() != AlphaComposite.SRC_OVER) {
            comp = AlphaComposite.SrcOver;
        }

        OGLRenderQueue rq = OGLRenderQueue.getInstance();
        rq.lock();
        try {
            validateOperation(sg2d, comp);

            RenderBuffer buf = rq.getBuffer();
            if (mask != null) {
                // we adjust the mask length so that the mask ends on a
                // 4-byte boundary
                int maskBytesRequired = (mask.length + 3) & (~3);
                int padding = maskBytesRequired - mask.length;
                int totalBytesRequired = 36 + maskBytesRequired;

                if (totalBytesRequired <= buf.capacity()) {
                    if (totalBytesRequired > buf.remaining()) {
                        // process the queue first and then enqueue the mask
                        rq.flushNow();
                    }
                    buf.putInt(MASK_FILL);
                    // enqueue parameters
                    buf.putInt(isSimpleColor ? 1 : 0);
                    buf.putInt(x).putInt(y).putInt(w).putInt(h);
                    buf.putInt(maskoff);
                    buf.putInt(maskscan);
                    buf.putInt(maskBytesRequired);
                    // enqueue the mask
                    buf.put(mask);
                    if (padding != 0) {
                        buf.position(buf.position() + padding);
                    }
                } else {
                    // queue is too small to accomodate entire mask; perform
                    // the operation directly on the queue flushing thread
                    rq.flushAndInvokeNow(new Runnable() {
                        public void run() {
                            maskFill(isSimpleColor, x, y, w, h,
                                     maskoff, maskscan, mask.length, mask);
                        }
                    });
                }
            } else {
                // this should work because the current fillpipe will always
                // be OGLRenderer for any of the compositing modes
                // for which OGLMaskFill is defined
                rq.ensureCapacity(20);
                buf.putInt(FILL_RECT);
                buf.putInt(x).putInt(y).putInt(w).putInt(h);
            }

            completeOperation();
        } finally {
            rq.unlock();
        }
    }

    private native void maskFill(boolean isColor,
                                 int x, int y, int w, int h,
                                 int maskoff, int maskscan, int masklen,
                                 byte[] mask);

    protected void validateOperation(SunGraphics2D sg2d,
                                     int pixel, Composite comp)
    {
        OGLSurfaceData dstData = (OGLSurfaceData)sg2d.surfaceData;
        OGLContext.validateContext(dstData, dstData,
                                   sg2d.getCompClip(), comp,
                                   null, pixel,
                                   OGLContext.NO_CONTEXT_FLAGS);
    }

    /**
     * Validates the state in the provided SunGraphics2D object and sets up
     * any special resources for this operation (e.g. enabling gradient
     * shading).
     */
    protected abstract void validateOperation(SunGraphics2D sg2d,
                                              Composite comp);

    /**
     * Completes the operation by releasing any resources that were set up
     * in validateOperation().
     */
    protected abstract void completeOperation();

    private static class Solid extends OGLMaskFill {
        protected Solid(SurfaceType srcType, CompositeType compType) {
            super(srcType, compType, true);
        }

        protected void validateOperation(SunGraphics2D sg2d, Composite comp) {
            validateOperation(sg2d, sg2d.pixel, comp);
        }

        protected void completeOperation() {
        }
    }

    private static class Gradient extends OGLMaskFill {
        protected Gradient(SurfaceType srcType, CompositeType compType) {
            super(srcType, compType, false);
        }

        protected void validateOperation(SunGraphics2D sg2d, Composite comp) {
            OGLRenderQueue rq = OGLRenderQueue.getInstance();
            // assert rq.lock.isHeldByCurrentThread();
            GradientPaint paint = (GradientPaint)sg2d.paint;
            int pixel = OGLContext.getExtraAlphaPixel(comp);
            validateOperation(sg2d, pixel, comp);
            OGLRenderer.Gradient.enableGradientPaint(rq, sg2d, paint);
        }

        protected void completeOperation() {
            OGLRenderQueue rq = OGLRenderQueue.getInstance();
            OGLRenderer.Gradient.disableGradientPaint(rq);
        }
    }

    private static class Texture extends OGLMaskFill {
        protected Texture(SurfaceType srcType, CompositeType compType) {
            super(srcType, compType, false);
        }

        protected void validateOperation(SunGraphics2D sg2d, Composite comp) {
            OGLRenderQueue rq = OGLRenderQueue.getInstance();
            // assert rq.lock.isHeldByCurrentThread();
            TexturePaint paint = (TexturePaint)sg2d.paint;
            int pixel = OGLContext.getExtraAlphaPixel(comp);
            validateOperation(sg2d, pixel, comp);
            OGLRenderer.Texture.enableTexturePaint(rq, sg2d, paint);
        }

        protected void completeOperation() {
            OGLRenderQueue rq = OGLRenderQueue.getInstance();
            OGLRenderer.Texture.disableTexturePaint(rq);
        }
    }
}
