/*
 * @(#)OGLRenderer.java	1.14 06/08/01
 *
 * Copyright 2006 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package sun.java2d.opengl;

import java.awt.Color;
import java.awt.Polygon;
import java.awt.Transparency;
import java.awt.GradientPaint;
import java.awt.TexturePaint;
import java.awt.geom.AffineTransform;
import java.awt.geom.Path2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;
import sun.awt.image.PixelConverter;
import sun.java2d.SunGraphics2D;
import sun.java2d.SurfaceData;
import sun.java2d.loops.CompositeType;
import sun.java2d.loops.GraphicsPrimitive;
import sun.java2d.pipe.BufferedRenderPipe;
import sun.java2d.pipe.RenderBuffer;
import sun.java2d.pipe.RenderQueue;
import sun.java2d.pipe.SpanIterator;
import static sun.java2d.pipe.BufferedOpCodes.*;

abstract class OGLRenderer extends BufferedRenderPipe {

    OGLRenderer(RenderQueue rq) {
        super(rq);
    }

    protected void validateOperation(SunGraphics2D sg2d,
                                     int pixel, boolean opaque)
    {
        int ctxflags =
            opaque ? OGLContext.SRC_IS_OPAQUE : OGLContext.NO_CONTEXT_FLAGS;
        OGLSurfaceData dstData = (OGLSurfaceData)sg2d.surfaceData;
        OGLContext.validateContext(dstData, dstData,
                                   sg2d.getCompClip(), sg2d.composite,
                                   null, pixel, ctxflags);
    }

    static class Solid extends OGLRenderer {

        Solid(RenderQueue rq) {
            super(rq);
        }

        protected void validateOperation(SunGraphics2D sg2d) {
            int pixel = sg2d.pixel;
            validateOperation(sg2d, pixel, (pixel >>> 24) == 0xff);
        }

        protected void completeOperation() {
        }
    }

    static class Gradient extends OGLRenderer {

        Gradient(RenderQueue rq) {
            super(rq);
        }

        /**
         * We use OpenGL's texture coordinate generator to automatically
         * apply a smooth gradient (either cyclic or acyclic) to the geometry
         * being rendered.  This technique is almost identical to the one
         * described in the comments for OGLRenderer.Texture.getContext(),
         * except the calculations take place in one dimension instead of two.
         * Instead of an anchor rectangle in the TexturePaint case, we use
         * the vector between the two GradientPaint end points in our
         * calculations.  The generator uses a single plane equation that
         * takes the (x,y) location (in device space) of the fragment being
         * rendered to calculate a (u) texture coordinate for that fragment:
         *     u = Ax + By + Cz + Dw
         *
         * The gradient renderer uses a two-pixel 1D texture where the first
         * pixel contains the first GradientPaint color, and the second pixel
         * contains the second GradientPaint color.  (Note that we repeat
         * those colors in the texture borders so that acyclic gradients
         * clamp the colors properly at the extremes.)  The following diagram
         * attempts to show the layout of the texture containing the two
         * GradientPaint colors (C1 and C2):
         *
         *                -----------------------------------
         *               |   C1   |   C1   |   C2   |   C2   |
         *               |(border)|        |        |(border)|
         *                -----------------------------------
         *                      u=0  .25  .5   .75  1
         *
         * We calculate our plane equation constants (A,B,D) such that u=0.25
         * corresponds to the first GradientPaint end point in user space and
         * u=0.75 corresponds to the second end point.  This is somewhat
         * non-obvious, but since the gradient colors are generated by
         * interpolating between C1 and C2, we want the pure color at the
         * end points, and we will get the pure color only when u correlates
         * to the center of a texel.  The following chart shows the expected
         * color for some sample values of u (where C' is the color halfway
         * between C1 and C2):
         *
         *       u value      acyclic (GL_CLAMP)      cyclic (GL_REPEAT)
         *       -------      ------------------      ------------------
         *        -0.25              C1                       C2
         *         0.0               C1                       C'
         *         0.25              C1                       C1
         *         0.5               C'                       C'
         *         0.75              C2                       C2
         *         1.0               C2                       C'
         *         1.25              C2                       C1
         *
         * Original inspiration for this technique came from UMD's Agile2D
         * project (GradientManager.java).
         */
        static void enableGradientPaint(RenderQueue rq,
                                        SunGraphics2D sg2d,
                                        GradientPaint paint)
        {
            // convert gradient colors to IntRgbaPre format
            PixelConverter pc = PixelConverter.RgbaPre.instance;
            Color c1 = paint.getColor1();
            int pixel1 = pc.rgbToPixel(c1.getRGB(), null);
            Color c2 = paint.getColor2();
            int pixel2 = pc.rgbToPixel(c2.getRGB(), null);

            // calculate plane equation constants
            AffineTransform at = (AffineTransform)sg2d.transform.clone();
            Point2D p = paint.getPoint1();
            double x = p.getX();
            double y = p.getY();
            at.translate(x, y);
            // now gradient point 1 is at the origin
            p = paint.getPoint2();
            x = p.getX() - x;
            y = p.getY() - y;
            double len = Math.sqrt(x * x + y * y);
            at.rotate(x, y);
            // now gradient point 2 is on the positive x-axis
            at.scale(2*len, 1);
            // now gradient point 2 is at (0.5, 0)
            at.translate(-0.25, 0);
            // now gradient point 1 is at (0.25, 0), point 2 is at (0.75, 0)

            double p0, p1, p3;
            try {
                at.invert();
                p0 = at.getScaleX();
                p1 = at.getShearX();
                p3 = at.getTranslateX();
            } catch (java.awt.geom.NoninvertibleTransformException e) {
                p0 = p1 = p3 = 0.0;
            }

            // assert rq.lock.isHeldByCurrentThread();
            rq.ensureCapacityAndAlignment(40, 8);
            RenderBuffer buf = rq.getBuffer();
            buf.putInt(ENABLE_GRADIENT_PAINT);
            buf.putInt(paint.isCyclic() ? 1 : 0);
            buf.putDouble(p0).putDouble(p1).putDouble(p3);
            buf.putInt(pixel1).putInt(pixel2);
        }

        static void disableGradientPaint(RenderQueue rq) {
            // assert rq.lock.isHeldByCurrentThread();
            rq.ensureCapacity(4);
            rq.getBuffer().putInt(DISABLE_GRADIENT_PAINT);
        }

        protected void validateOperation(SunGraphics2D sg2d) {
            GradientPaint paint = (GradientPaint)sg2d.paint;
            boolean opaque = (paint.getTransparency() == Transparency.OPAQUE);
            int pixel = OGLContext.getExtraAlphaPixel(sg2d.composite);
            validateOperation(sg2d, pixel, opaque);
            enableGradientPaint(rq, sg2d, paint);
        }

        protected void completeOperation() {
            disableGradientPaint(rq);
        }
    }

    static class Texture extends OGLRenderer {

        Texture(RenderQueue rq) {
            super(rq);
        }

        /**
         * Returns true if the given TexturePaint instance can be used by the
         * accelerated OGLRenderer.Texture implementation.  A TexturePaint is
         * considered valid if the following conditions are met:
         *   - the texture image dimensions are power-of-two (or the
         *     GL_ARB_texture_non_power_of_two extension is present)
         *   - the texture image can be (or is already) cached in an OpenGL
         *     texture object
         */
        static boolean isPaintValid(SunGraphics2D sg2d,
                                    TexturePaint paint)
        {
            OGLSurfaceData dstData = (OGLSurfaceData)sg2d.surfaceData;
            BufferedImage bi = paint.getImage();

            // see if texture-non-pow2 extension is available
            if (!dstData.isTexNonPow2Available()) {
                int imgw = bi.getWidth();
                int imgh = bi.getHeight();

                // verify that the texture image dimensions are pow2
                if ((imgw & (imgw - 1)) != 0 || (imgh & (imgh - 1)) != 0) {
                    return false;
                }
            }

            SurfaceData srcData =
                SurfaceData.getSourceSurfaceData(bi, dstData,
                                                 CompositeType.SrcOver,
                                                 null, false);
            if (!(srcData instanceof OGLSurfaceData)) {
                // REMIND: this is a hack that attempts to cache the system
                //         memory image from the TexturePaint instance into an
                //         OpenGL texture...
                srcData =
                    SurfaceData.getSourceSurfaceData(bi, dstData,
                                                     CompositeType.SrcOver,
                                                     null, false);
                if (!(srcData instanceof OGLSurfaceData)) {
                    return false;
                }
            }

            // verify that the source surface is actually a texture
            OGLSurfaceData oglData = (OGLSurfaceData)srcData;
            if (oglData.getType() != OGLSurfaceData.TEXTURE) {
                return false;
            }

            return true;
        }

        /**
         * We use OpenGL's texture coordinate generator to automatically
         * map the TexturePaint image to the geometry being rendered.  The
         * generator uses two separate plane equations that take the (x,y)
         * location (in device space) of the fragment being rendered to
         * calculate (u,v) texture coordinates for that fragment:
         *     u = Ax + By + Cz + Dw
         *     v = Ex + Fy + Gz + Hw
         *
         * Since we use a 2D orthographic projection, we can assume that z=0
         * and w=1 for any fragment.  So we need to calculate appropriate
         * values for the plane equation constants (A,B,D) and (E,F,H) such
         * that {u,v}=0 for the top-left of the TexturePaint's anchor
         * rectangle and {u,v}=1 for the bottom-right of the anchor rectangle.
         * We can easily make the texture image repeat for {u,v} values
         * outside the range [0,1] by specifying the GL_REPEAT texture wrap
         * mode.
         *
         * Calculating the plane equation constants is surprisingly simple.
         * We can think of it as an inverse matrix operation that takes
         * device space coordinates and transforms them into user space
         * coordinates that correspond to a location relative to the anchor
         * rectangle.  First, we translate and scale the current user space
         * transform by applying the anchor rectangle bounds.  We then take
         * the inverse of this affine transform.  The rows of the resulting
         * inverse matrix correlate nicely to the plane equation constants
         * we were seeking.
         */
        static void enableTexturePaint(RenderQueue rq,
                                       SunGraphics2D sg2d,
                                       TexturePaint paint)
        {
            BufferedImage bi = paint.getImage();
            SurfaceData dstData = sg2d.surfaceData;
            SurfaceData srcData =
                SurfaceData.getSourceSurfaceData(bi, dstData,
                                                 CompositeType.SrcOver,
                                                 null, false);
            if (!(srcData instanceof OGLSurfaceData)) {
                // REMIND: return error value?
                return;
            }

            boolean filter =
                (sg2d.interpolationType !=
                 AffineTransformOp.TYPE_NEAREST_NEIGHBOR);

            // calculate plane equation constants
            AffineTransform at = (AffineTransform)sg2d.transform.clone();
            Rectangle2D anchor = paint.getAnchorRect();
            at.translate(anchor.getX(), anchor.getY());
            at.scale(anchor.getWidth(), anchor.getHeight());

            double xp0, xp1, xp3, yp0, yp1, yp3;
            try {
                at.invert();
                xp0 = at.getScaleX();
                xp1 = at.getShearX();
                xp3 = at.getTranslateX();
                yp0 = at.getShearY();
                yp1 = at.getScaleY();
                yp3 = at.getTranslateY();
            } catch (java.awt.geom.NoninvertibleTransformException e) {
                xp0 = xp1 = xp3 = yp0 = yp1 = yp3 = 0.0;
            }

            long pSrcOps = srcData.getNativeOps();

            // assert rq.lock.isHeldByCurrentThread();
            rq.ensureCapacityAndAlignment(64, 8);
            RenderBuffer buf = rq.getBuffer();
            buf.putInt(ENABLE_TEXTURE_PAINT);
            buf.putInt(filter ? 1 : 0);
            buf.putLong(pSrcOps);
            buf.putDouble(xp0).putDouble(xp1).putDouble(xp3);
            buf.putDouble(yp0).putDouble(yp1).putDouble(yp3);
        }

        static void disableTexturePaint(RenderQueue rq) {
            // assert rq.lock.isHeldByCurrentThread();
            rq.ensureCapacity(4);
            rq.getBuffer().putInt(DISABLE_TEXTURE_PAINT);
        }

        protected void validateOperation(SunGraphics2D sg2d) {
            TexturePaint paint = (TexturePaint)sg2d.paint;
            boolean opaque = (paint.getTransparency() == Transparency.OPAQUE);
            int pixel = OGLContext.getExtraAlphaPixel(sg2d.composite);
            validateOperation(sg2d, pixel, opaque);
            enableTexturePaint(rq, sg2d, paint);
        }

        protected void completeOperation() {
            disableTexturePaint(rq);
        }
    }

    void copyArea(SunGraphics2D sg2d,
                  int x, int y, int w, int h, int dx, int dy)
    {
        rq.lock();
        try {
            boolean opaque =
                sg2d.surfaceData.getTransparency() == Transparency.OPAQUE;
            validateOperation(sg2d, sg2d.pixel, opaque);
            rq.ensureCapacity(28);
            buf.putInt(COPY_AREA);
            buf.putInt(x).putInt(y).putInt(w).putInt(h);
            buf.putInt(dx).putInt(dy);
            completeOperation();
        } finally {
            rq.unlock();
        }
    }

    protected native void drawPoly(int[] xPoints, int[] yPoints,
                                   int nPoints, boolean isClosed,
                                   int transX, int transY);

    OGLRenderer traceWrap() {
	return new Tracer(this);
    }

    private class Tracer extends OGLRenderer {
        private OGLRenderer oglr;
        Tracer(OGLRenderer oglr) {
            super(oglr.rq);
            this.oglr = oglr;
        }
        protected void validateOperation(SunGraphics2D sg2d) {
            oglr.validateOperation(sg2d);
        }
        protected void completeOperation() {
            oglr.completeOperation();
        }
        public void drawLine(SunGraphics2D sg2d,
                             int x1, int y1, int x2, int y2)
        {
	    GraphicsPrimitive.tracePrimitive("OGLDrawLine");
	    oglr.drawLine(sg2d, x1, y1, x2, y2);
	}
        public void drawRect(SunGraphics2D sg2d, int x, int y, int w, int h) {
	    GraphicsPrimitive.tracePrimitive("OGLDrawRect");
	    oglr.drawRect(sg2d, x, y, w, h);
	}
        protected void drawPoly(SunGraphics2D sg2d,
                                int[] xPoints, int[] yPoints,
                                int nPoints, boolean isClosed)
        {
	    GraphicsPrimitive.tracePrimitive("OGLDrawPoly");
	    oglr.drawPoly(sg2d, xPoints, yPoints, nPoints, isClosed);
	}
        public void fillRect(SunGraphics2D sg2d, int x, int y, int w, int h) {
	    GraphicsPrimitive.tracePrimitive("OGLFillRect");
	    oglr.fillRect(sg2d, x, y, w, h);
	}
        protected void drawPath(SunGraphics2D sg2d,
                                Path2D.Float p2df, int transx, int transy)
        {
            GraphicsPrimitive.tracePrimitive("OGLDrawPath");
            oglr.drawPath(sg2d, p2df, transx, transy);
        }
        protected void fillPath(SunGraphics2D sg2d,
                                Path2D.Float p2df, int transx, int transy)
        {
            GraphicsPrimitive.tracePrimitive("OGLFillPath");
            oglr.fillPath(sg2d, p2df, transx, transy);
        }
        protected void fillSpans(SunGraphics2D sg2d, SpanIterator si,
                                 int transx, int transy)
	{
            GraphicsPrimitive.tracePrimitive("OGLFillSpans");
            oglr.fillSpans(sg2d, si, transx, transy);
	}
        public void copyArea(SunGraphics2D sg2d,
                             int x, int y, int w, int h, int dx, int dy)
        {
            GraphicsPrimitive.tracePrimitive("OGLCopyArea");
            oglr.copyArea(sg2d, x, y, w, h, dx, dy);
	}
    }
}
