/*
 * @(#)X11RemoteOffScreenImage.java	1.24 05/11/17
 *
 * Copyright 2006 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package sun.java2d.x11;

import java.awt.Component;
import java.awt.GraphicsConfiguration;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.WritableRaster;
import sun.awt.X11ComponentPeer;
import sun.awt.X11GraphicsConfig;
import sun.awt.image.OffScreenSurfaceManager;
import sun.awt.image.RemoteOffScreenImage;
import sun.awt.image.SurfaceManager;
import sun.java2d.SurfaceData;
import sun.java2d.opengl.GLXSurfaceData;
import sun.java2d.opengl.GLXGraphicsConfig;
import sun.java2d.opengl.OGLSurfaceData;

/**
 * This class extends the functionality of RemoteOffScreenImage by
 * forcing the use of a pixmap-based SurfaceData object.  In
 * X11CachingSurfaceManager, the default SurfaceData object is the default
 * BufImgSurfaceData, setup by a call to the superclass
 * constructor.  Acceleration is achieved by caching a version of
 * the data in a pixmap and copying from that pixmap when appropriate.
 * <P>
 * This is sufficient for accelerating sprite-type images, which are
 * rendered to infrequently but copied from often.  Back buffers 
 * do not benefit from this acceleration since the cached copy
 * would never be used.  This is deemed sufficient for local X
 * usage; users performing complex Java2D operations (e.g., 
 * anti-aliasing, compositing, text, or anything that requires
 * read-modify-write operations) would see a degradation in performance
 * were we to force all operations to go through the pixmap-based
 * image.
 * <P>
 * In the RemoteX case, however, performance is so abysmal in the
 * case of double-buffering (with the back buffer living in the 
 * Java heap), that the advantage of speeding up the basic 
 * applications (i.e., those using 1.1 API or simple Swing operations,
 * and thus avoiding complex Java2D read-modify-write operations)
 * is judged to outweigh the possible performance loss in some
 * applications from having the back buffer located in a pixmap.
 * <P>
 * The decision to instantiate this class (versus OffScreenImage)
 * is based on whether we are running remotely and whether the
 * user has enabled/disabled a property related to this issue:
 * -Dsun.java2d.pmoffscreen
 */
public class X11RemoteOffScreenImage extends RemoteOffScreenImage {

    public X11RemoteOffScreenImage(Component c,
                                   ColorModel cm, WritableRaster raster,
				   boolean isRasterPremultiplied)
    {
	super(c, cm, raster, isRasterPremultiplied);
        createNativeRaster();
    }

    protected SurfaceManager createSurfaceManager() {
        return new X11RemoteSurfaceManager(this);
    }

    private class X11RemoteSurfaceManager extends OffScreenSurfaceManager {
        public X11RemoteSurfaceManager(BufferedImage bi) {
            super(c, bi);
        }

	protected boolean isDestSurfaceAccelerated(SurfaceData destSD) {
	    return (destSD instanceof GLXSurfaceData || 
		    (destSD instanceof X11SurfaceData &&
		     X11SurfaceData.isAccelerationEnabled()));
	}

	protected SurfaceData createAccelSurface(GraphicsConfiguration gc, 
						 int width, int height) 
	{
	    if (gc instanceof GLXGraphicsConfig) {
		return GLXSurfaceData.createData((GLXGraphicsConfig)gc,
						 width, height,
						 gc.getColorModel(), bImg,
						 OGLSurfaceData.TEXTURE);
	    } else {
		return X11SurfaceData.createData((X11GraphicsConfig)gc,
						 width, height,
						 gc.getColorModel(),
						 bImg, 0, 0);
	    }
	}
    }
}
