/*
 * @(#)P11KeyGenerator.java	1.5 05/11/17
 *
 * Copyright 2006 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package sun.security.pkcs11;

import java.security.*;
import java.security.spec.AlgorithmParameterSpec;

import javax.crypto.*;

import static sun.security.pkcs11.TemplateManager.*;
import sun.security.pkcs11.wrapper.*;
import static sun.security.pkcs11.wrapper.PKCS11Constants.*;

/**
 * KeyGenerator implementation class. This class currently supports 
 * DES, DESede, AES, ARCFOUR, and Blowfish.
 *
 * @author  Andreas Sterbenz
 * @version 1.5, 11/17/05
 * @since   1.5
 */
final class P11KeyGenerator extends KeyGeneratorSpi {
    
    // token instance
    private final Token token;
    
    // algorithm name
    private final String algorithm;
    
    // mechanism id
    private long mechanism;
    
    // raw key size in bits, e.g. 64 for DES. Always valid.
    private int keySize;
    
    // bits of entropy in the key, e.g. 56 for DES. Always valid.
    private int significantKeySize;
    
    // keyType (CKK_*), neede for TemplateManager call only
    private long keyType;
    
    P11KeyGenerator(Token token, String algorithm, long mechanism) 
	    throws PKCS11Exception {
	super();
	this.token = token;
	this.algorithm = algorithm;
	this.mechanism = mechanism;
	setDefaultKeySize();
    }
    
    // set default keysize and also initialize keyType
    private void setDefaultKeySize() {
	switch ((int)mechanism) {
	case (int)CKM_DES_KEY_GEN:
	    keySize = 64;
	    significantKeySize = 56;
	    keyType = CKK_DES;
	    break;
	case (int)CKM_DES2_KEY_GEN:
	    keySize = 128;
	    significantKeySize = 112;
	    keyType = CKK_DES2;
	    break;
	case (int)CKM_DES3_KEY_GEN:
	    keySize = 192;
	    significantKeySize = 168;
	    keyType = CKK_DES3;
	    break;
	case (int)CKM_AES_KEY_GEN:
	    keyType = CKK_AES;
	    keySize = 128;
	    significantKeySize = 128;
	    break;
	case (int)CKM_RC4_KEY_GEN:
	    keyType = CKK_RC4;
	    keySize = 128;
	    significantKeySize = 128;
	    break;
	case (int)CKM_BLOWFISH_KEY_GEN:
	    keyType = CKK_BLOWFISH;
	    keySize = 128;
	    significantKeySize = 128;
	    break;
	default:
	    throw new ProviderException("Unknown mechanism " + mechanism);
	}
    }

    // see JCE spec
    protected void engineInit(SecureRandom random) {
	token.ensureValid();
	setDefaultKeySize();
    }

    // see JCE spec
    protected void engineInit(AlgorithmParameterSpec params, 
	    SecureRandom random) throws InvalidAlgorithmParameterException {
	throw new InvalidAlgorithmParameterException
		("AlgorithmParameterSpec not supported");
    }

    // see JCE spec
    protected void engineInit(int keySize, SecureRandom random) {
	token.ensureValid();
	switch ((int)mechanism) {
	case (int)CKM_DES_KEY_GEN:
	    if (keySize != 64) {
		throw new InvalidParameterException
			("DES key length must be 64 bits");
	    }
	    this.keySize = 64;
	    significantKeySize = 56;
	    break;
	case (int)CKM_DES2_KEY_GEN:
	case (int)CKM_DES3_KEY_GEN:
	    // XXX check mech for requested key length is available
	    if (keySize == 128) {
		mechanism = CKM_DES2_KEY_GEN;
		this.keySize = 128;
		significantKeySize = 112;
	    } else if (keySize == 192) {
		mechanism = CKM_DES3_KEY_GEN;
		this.keySize = 192;
		significantKeySize = 168;
	    } else {
		throw new InvalidParameterException
			("DESede key length must be 128 or 192 bits");
	    }
	    break;
	case (int)CKM_AES_KEY_GEN:
	    if ((keySize != 128) && (keySize != 192) && (keySize != 256)) {
		throw new InvalidParameterException
			("AES key length must be 128, 192, or 256 bits");
	    }
	    this.keySize = keySize;
	    significantKeySize = keySize;
	    break;
	case (int)CKM_RC4_KEY_GEN:
	    if ((keySize < 40) || (keySize > 1024)) {
		throw new InvalidParameterException
		    ("ARCFOUR key length must be between 40 and 1024 bits");
	    }
	    this.keySize = keySize;
	    significantKeySize = keySize;
	    break;
	case (int)CKM_BLOWFISH_KEY_GEN:
	    if ((keySize < 40) || (keySize > 448)) {
		throw new InvalidParameterException
		    ("Blowfish key length must be between 40 and 448 bits");
	    }
	    this.keySize = keySize;
	    significantKeySize = keySize;
	    break;
	default:
	    throw new ProviderException("Unknown mechanism " + mechanism);
	}
    }

    // see JCE spec
    protected SecretKey engineGenerateKey() {
	Session session = null;
	try {
	    session = token.getObjSession();
	    CK_ATTRIBUTE[] attributes;
	    switch ((int)keyType) {
	    case (int)CKK_DES:
	    case (int)CKK_DES2:
	    case (int)CKK_DES3:
	        // fixed length, do not specify CKA_VALUE_LEN
		attributes = new CK_ATTRIBUTE[] {
		    new CK_ATTRIBUTE(CKA_CLASS, CKO_SECRET_KEY),
		};
		break;
	    default:
		attributes = new CK_ATTRIBUTE[] {
		    new CK_ATTRIBUTE(CKA_CLASS, CKO_SECRET_KEY),
		    new CK_ATTRIBUTE(CKA_VALUE_LEN, keySize >> 3),
		};
		break;
	    }
	    attributes = token.getAttributes
	    	(O_GENERATE, CKO_SECRET_KEY, keyType, attributes);
	    long keyID = token.p11.C_GenerateKey
	    	(session.id(), new CK_MECHANISM(mechanism), attributes);
	    return P11Key.secretKey
	    	(session, keyID, algorithm, significantKeySize, attributes);
	} catch (PKCS11Exception e) {
	    throw new ProviderException("Could not generate key", e);
	} finally {
	    token.releaseSession(session);
	}
    }

}

